/***********************************************************************
 * File: utf8.c
 * 	(Extracted from
 *	   "$Id: utf8.c,v 1.2 2008/10/06 13:03:26 maoyam Exp maoyam $")
 *
 * Abstracts:
 *  This C source implements a character converter between UTF-8 and
 *  EUC-JP/SJIS.
 *
 * Copyright (C) 2008 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/


/* include this library definitions */
#include "utf8.h"
#include <iconv.h>

/* this function sanitizes a string (safer UTF-8 encoding)
 * EINVAL means "illegal arguments".
 */
unsigned char *sanitize_for_utf8(Sanitize_for_UTF8_e	str_encoding
								, size_t				length
								, unsigned char			*str
								, char					replace_code
								)
{
	static struct garbled_characters_t {
		unsigned char	*from, *to; /* set range of garbled code region */
	}
	  eucjp_garbled[] = {
		  { /* JIS X 0201 (Kana) 21 to 5F */
			(unsigned char *)"\x8e\x21", (unsigned char *)"\x8e\x5f"
			}
		, {	/* JIS X 0208 (level 1) 02-15 to 02-25 */
			(unsigned char *)"\xa2\xaf", (unsigned char *)"\xa2\xb9"
			}
		, {	/* JIS X 0208 (level 1) 02-34 to 02-41 */
			(unsigned char *)"\xa2\xc2", (unsigned char *)"\xa2\xc9"
			}
		, {	/* JIS X 0208 (level 1) 02-49 to 02-59 */
			(unsigned char *)"\xa2\xd1", (unsigned char *)"\xa2\xdb"
			}
		, {	/* JIS X 0208 (level 1) 02-75 to 02-81 */
			(unsigned char *)"\xa2\xeb", (unsigned char *)"\xa2\xf1"
			}
		, {	/* JIS X 0208 (level 1) 02-90 to 02-93 */
			(unsigned char *)"\xa2\xfa", (unsigned char *)"\xa2\xfd"
			}
		, {	/* JIS X 0208 (level 1) 03-01 to 03-15 */
			(unsigned char *)"\xa3\xa1", (unsigned char *)"\xa3\xaf"
			}
		, {	/* JIS X 0208 (level 1) 03-26 to 03-32 */
			(unsigned char *)"\xa3\xba", (unsigned char *)"\xa3\xc0"
			}
		, {	/* JIS X 0208 (level 1) 03-59 to 03-64 */
			(unsigned char *)"\xa3\xdb", (unsigned char *)"\xa3\xe0"
			}
		, {	/* JIS X 0208 (level 1) 03-91 to 03-94 */
			(unsigned char *)"\xa3\xfb", (unsigned char *)"\xa3\xfe"
			}
		, {	/* JIS X 0208 (level 1) 04-84 to 04-94 */
			(unsigned char *)"\xa4\xf4", (unsigned char *)"\xa4\xfe"
			}
		, {	/* JIS X 0208 (level 1) 05-87 to 05-94 */
			(unsigned char *)"\xa5\xf7", (unsigned char *)"\xa5\xfe"
			}
		, {	/* JIS X 0208 (level 1) 06-25 to 06-32 */
			(unsigned char *)"\xa6\xb9", (unsigned char *)"\xa6\xc0"
			}
		, {	/* JIS X 0208 (level 1) 06-57 to 06-94 */
			(unsigned char *)"\xa6\xd9", (unsigned char *)"\xa6\xfe"
			}
		, {	/* JIS X 0208 (level 1) 07-34 to 07-47 */
			(unsigned char *)"\xa7\xc2", (unsigned char *)"\xa7\xd0"
			}
		, {	/* JIS X 0208 (level 1) 07-82 to 07-94 */
			(unsigned char *)"\xa7\xf2", (unsigned char *)"\xa7\xfe"
			}
		, {	/* JIS X 0208 (level 1) 08-33 to 08-94 */
			(unsigned char *)"\xa8\xc1", (unsigned char *)"\xa8\xfe"
			}
		, {	/* JIS X 0208 (level 1) 09-01 to 09-94 */
			(unsigned char *)"\xa9\xa1", (unsigned char *)"\xa9\xfe"
			}
		, {	/* JIS X 0208 (level 1) 10-01 to 10-94 */
			(unsigned char *)"\xaa\xa1", (unsigned char *)"\xaa\xfe"
			}
		, {	/* JIS X 0208 (level 1) 11-01 to 11-94 */
			(unsigned char *)"\xab\xa1", (unsigned char *)"\xab\xfe"
			}
		, {	/* JIS X 0208 (level 1) 12-01 to 12-94 */
			(unsigned char *)"\xac\xa1", (unsigned char *)"\xac\xfe"
			}
		, {	/* JIS X 0208 (level 1) 13-01 to 13-94 (circled digits, etc) */
			(unsigned char *)"\xad\xa1", (unsigned char *)"\xad\xfe"
			}
		, {	/* JIS X 0208 (level 1) 14-01 to 14-94 */
			(unsigned char *)"\xae\xa1", (unsigned char *)"\xae\xfe"
			}
		, {	/* JIS X 0208 (level 1) 15-01 to 15-94 */
			(unsigned char *)"\xaf\xa1", (unsigned char *)"\xaf\xfe"
			}
		, {	/* JIS X 0208 (level 1) 47-52 to 47-94 */
			(unsigned char *)"\xcf\xd4", (unsigned char *)"\xcf\xfe"
			}
		, {	/* JIS X 0208 (level 2) 84-07 to 84-94 */
			(unsigned char *)"\xf4\xa7", (unsigned char *)"\xf4\xfe"
			}
		, {	/* JIS X 0208 (level 2) 85-01 to 85-94 */
			(unsigned char *)"\xf5\xa1", (unsigned char *)"\xf5\xfe"
			}
		, {	/* JIS X 0208 (level 2) 86-01 to 86-94 (JIS not assigned) */
			(unsigned char *)"\xf6\xa1", (unsigned char *)"\xf6\xfe"
			}
		, {	/* JIS X 0208 (level 2) 87-01 to 87-94 (JIS not assigned) */
			(unsigned char *)"\xf7\xa1", (unsigned char *)"\xf7\xfe"
			}
		, {	/* JIS X 0208 (level 2) 88-01 to 88-94 (JIS not assigned) */
			(unsigned char *)"\xf8\xa1", (unsigned char *)"\xf8\xfe"
			}
		, {	/* JIS X 0208 (level 2) 89-01 to 89-94 (debian garbled) */
			(unsigned char *)"\xf9\xa1", (unsigned char *)"\xf9\xfe"
			}
		, {	/* JIS X 0208 (level 2) 90-01 to 90-94 (debian garbled) */
			(unsigned char *)"\xfa\xa1", (unsigned char *)"\xfa\xfe"
			}
		, {	/* JIS X 0208 (level 2) 91-01 to 91-94 (debian garbled) */
			(unsigned char *)"\xfb\xa1", (unsigned char *)"\xfb\xfe"
			}
		, {	/* JIS X 0208 (level 2) 92-01 to 92-94 (debian garbled) */
			(unsigned char *)"\xfc\xa1", (unsigned char *)"\xfc\xfe"
			}
		, {	/* JIS X 0208 (level 2) 93-01 to 93-94 (JIS not assigned) */
			(unsigned char *)"\xfd\xa1", (unsigned char *)"\xfd\xfe"
			}
		, {	/* JIS X 0208 (level 2) 94-01 to 94-94 (JIS not assigned) */
			(unsigned char *)"\xfe\xa1", (unsigned char *)"\xfe\xfe"
			}
		, {	/* end of `garbled' table */	NULL,	NULL	}
		}
	, sjis_garbled[] = {
		  {	/* JIS X 0208 (level 1) 02-15 to 02-25 */
			(unsigned char *)"\x81\xad", (unsigned char *)"\x81\xb7"
			}
		, {	/* JIS X 0208 (level 1) 02-34 to 02-41 */
			(unsigned char *)"\x81\xc0", (unsigned char *)"\x81\xc7"
			}
		, {	/* JIS X 0208 (level 1) 02-49 to 02-59 */
			(unsigned char *)"\x81\xcf", (unsigned char *)"\x81\xd9"
			}
		, {	/* JIS X 0208 (level 1) 02-75 to 02-81 */
			(unsigned char *)"\x81\xe9", (unsigned char *)"\x81\xef"
			}
		, {	/* JIS X 0208 (level 1) 02-90 to 02-93 */
			(unsigned char *)"\x81\xf8", (unsigned char *)"\x81\xfb"
			}
		, {	/* JIS X 0208 (level 1) 03-01 to 03-15 */
			(unsigned char *)"\x82\x40", (unsigned char *)"\x82\x4e"
			}
		, {	/* JIS X 0208 (level 1) 03-26 to 03-32 */
			(unsigned char *)"\x82\x59", (unsigned char *)"\x82\x5f"
			}
		, {	/* JIS X 0208 (level 1) 03-59 to 03-64 */
			(unsigned char *)"\x82\x7a", (unsigned char *)"\x82\x80"
			}
		, {	/* JIS X 0208 (level 1) 03-91 to 03-94 */
			(unsigned char *)"\x82\x9b", (unsigned char *)"\x82\x9e"
			}
		, {	/* JIS X 0208 (level 1) 04-84 to 04-94 */
			(unsigned char *)"\x82\xf2", (unsigned char *)"\x82\xfc"
			}
		, {	/* JIS X 0208 (level 1) 05-87 to 05-94 */
			(unsigned char *)"\x83\x97", (unsigned char *)"\x83\x9e"
			}
		, {	/* JIS X 0208 (level 1) 06-25 to 06-32 */
			(unsigned char *)"\x83\xb7", (unsigned char *)"\x83\xbe"
			}
		, {	/* JIS X 0208 (level 1) 06-57 to 06-94 */
			(unsigned char *)"\x83\xd7", (unsigned char *)"\x83\xfc"
			}
		, {	/* JIS X 0208 (level 1) 07-34 to 07-48 */
			(unsigned char *)"\x84\x61", (unsigned char *)"\x84\x6f"
			}
		, {	/* JIS X 0208 (level 1) 07-82 to 07-94 */
			(unsigned char *)"\x84\x92", (unsigned char *)"\x84\x9e"
			}
		, {	/* JIS X 0208 (level 1) 08-33 to 08-94 */
			(unsigned char *)"\x84\xbf", (unsigned char *)"\x84\xfc"
			}
		, {	/* JIS X 0208 (level 1) 09-01 to 09-94 */
			(unsigned char *)"\x85\x40", (unsigned char *)"\x85\x9e"
			}
		, {	/* JIS X 0208 (level 1) 10-01 to 10-94 */
			(unsigned char *)"\x85\x9f", (unsigned char *)"\x85\xfc"
			}
		, {	/* JIS X 0208 (level 1) 11-01 to 11-94 */
			(unsigned char *)"\x86\x40", (unsigned char *)"\x86\x9e"
			}
		, {	/* JIS X 0208 (level 1) 12-01 to 12-94 */
			(unsigned char *)"\x86\x9f", (unsigned char *)"\x86\xfc"
			}
		, {	/* JIS X 0208 (level 1) 13-01 to 13-94 (circled digits, etc) */
			(unsigned char *)"\x87\x40", (unsigned char *)"\x87\x9e"
			}
		, {	/* JIS X 0208 (level 1) 14-01 to 14-94 */
			(unsigned char *)"\x87\x9f", (unsigned char *)"\x87\xfc"
			}
		, {	/* JIS X 0208 (level 1) 15-01 to 15-94 */
			(unsigned char *)"\x88\x40", (unsigned char *)"\x88\x9e"
			}
		, {	/* JIS X 0208 (level 1) 47-52 to 47-94 */
			(unsigned char *)"\x98\x73", (unsigned char *)"\x98\x9e"
			}
		, {	/* JIS X 0208 (level 2) 84-07 to 84-94 */
			(unsigned char *)"\xea\xa5", (unsigned char *)"\xea\xfc"
			}
		, {	/* JIS X 0208 (level 2) 85-01 to 85-94 */
			(unsigned char *)"\xeb\x40", (unsigned char *)"\xeb\x9e"
			}
		, {	/* JIS X 0208 (level 2) 86-01 to 86-94 (JIS not assigned) */
			(unsigned char *)"\xeb\x9f", (unsigned char *)"\xeb\xfc"
			}
		, {	/* JIS X 0208 (level 2) 87-01 to 87-94 (JIS not assigned) */
			(unsigned char *)"\xec\x40", (unsigned char *)"\xec\x9e"
			}
		, {	/* JIS X 0208 (level 2) 88-01 to 88-94 (JIS not assigned) */
			(unsigned char *)"\xec\x9f", (unsigned char *)"\xec\xfc"
			}
		, {	/* JIS X 0208 (level 2) 89-01 to 89-94 (debian garbled) */
			(unsigned char *)"\xed\x40", (unsigned char *)"\xed\x9e"
			}
		, {	/* JIS X 0208 (level 2) 90-01 to 90-94 (debian garbled) */
			(unsigned char *)"\xed\x9f", (unsigned char *)"\xed\xfc"
			}
		, {	/* JIS X 0208 (level 2) 91-01 to 91-94 (debian garbled) */
			(unsigned char *)"\xee\x40", (unsigned char *)"\xee\x9e"
			}
		, {	/* JIS X 0208 (level 2) 92-01 to 92-94 (debian garbled) */
			(unsigned char *)"\xee\x9f", (unsigned char *)"\xee\xfc"
			}
		, {	/* JIS X 0208 (level 2) 93-01 to 93-94 (JIS not assigned) */
			(unsigned char *)"\xef\x40", (unsigned char *)"\xef\x9e"
			}
		, {	/* JIS X 0208 (level 2) 94-01 to 94-94 (JIS not assigned) */
			(unsigned char *)"\xef\x9f", (unsigned char *)"\xef\xfc"
			}
		, {	/* end of `garbled' table */	NULL,	NULL	}
		};
	struct garbled_characters_t	*garbled, *gap;
	unsigned char				*result = str;
	size_t						size;

	if ((((str_encoding < Sanitize_for_UTF8_e_EUCJP
	|| str_encoding >= Sanitize_for_UTF8_e_N)
	|| str == NULL)
	|| replace_code <= (char)'\0')
	) {	/* ERROR: illegal argument */
		errno	= EINVAL	/* invalid arguments */
		; return	(unsigned char *)NULL
		;
	}

	/* set table of `garbled' code */
	garbled	= ((str_encoding == Sanitize_for_UTF8_e_EUCJP)? eucjp_garbled
				:											sjis_garbled
				);
	/* process the source string... */
	while ((length > 0) && *str != (unsigned char)'\0') {
		if (*str <= (unsigned char)0x7f) {
			/* leave the ASCII (0x01 to 0x7d, 0x7f) and C0-control codes */
			++str; --length; continue;
		}

		/* process extra codes (each source encoding) */
		if (str_encoding == Sanitize_for_UTF8_e_EUCJP) {
			if (*str == (unsigned char)0x8f) {
				/* EUC-JP extra code (JIS X 0212) needs three bytes:
				 * => replace it, using `replace_code (ASCII code)'
				 */
				size	= ((length < 3)? length: 3)
				; memset(str, (int)replace_code, size)
				; str += size; length -= size
				; continue
				;
			}
		}
		else if (str_encoding == Sanitize_for_UTF8_e_SJIS) {
			if ((*str >= (unsigned char)0xa1)
			&& *str <= (unsigned char)0xdf) {
				/* replace the SJIS hankaku-kana code (single byte) */
				*str = (unsigned char)replace_code
				; ++str; --length; continue
				;
			}
		}

		/* check left length */
		if (length < 2) {
			/* process the last one byte */
			if (*str > (unsigned char)0x7f) {
				/* replace the non-ASCII code. */
				*str	= (unsigned char)replace_code;
			}
			++str; --length; continue;
		}

		/* check and process the two bytes... */
		for (gap = garbled, size = 2
		; (gap->from != NULL) && gap->to != NULL
		; ++gap
		) {
			if ((memcmp(str, gap->from, size) >= 0)
			&& memcmp(str, gap->to, size) <= 0) {
				break;	/* find garbled code */
			}
		}
		if ((gap->from != NULL) && gap->to != NULL) {
			/* replace garbled code, using `replace_code (ASCII code)' */
			memset(str, (int)replace_code, size);
		}
		str += size; length -= size;
	}

	return	result;	/* processed normally */
}

/* convert the EUC-JP or SJIS encoded string to UTF-8 encoded.
 * EINVAL means "illegal arguments".
 * EPERM means "system not support EUC-JP, SJIS or UTF-8 encode".
 * EILSEQ means "`inbuf' contains illegal character".
 */
unsigned char *to_utf8(Sanitize_for_UTF8_e	inbuf_encode
						, unsigned char		*inbuf
						, size_t			inbuf_length
						, char				replace_code
						, unsigned char		*outbuf
						, size_t			outbuf_size
						, char				padding_code
						, unsigned long		how
						)
{
	unsigned char	*dup_inbuf = NULL, *inbufp = NULL
					, *iconved_buf = NULL, *outbufp = NULL
					, *result = NULL
					;
	iconv_t			cd =(iconv_t)(-1);
	size_t			inbytesleft = inbuf_length, outbytesleft = outbuf_size;
	char			*tocode = NULL, *fromcode = NULL;
	int				sv_errno = 0;

	if (((((inbuf_encode < Sanitize_for_UTF8_e_EUCJP)
	|| inbuf_encode >= Sanitize_for_UTF8_e_N)
	|| inbuf == NULL)
	|| outbuf == NULL)
	) {	/* Error: argument is illegal. */
		errno	= EINVAL;	return	NULL; /* invalid arguments */
	}

	if (replace_code == '\0') {
		replace_code	= ' ';	/* default replace_code is space (0x20) */
	}

	/* allocate and initialize the iconv-ed buffer */
	if ((iconved_buf = malloc(outbuf_size + 1)) == NULL) {
		/* Error: can't allocate iconv-ed string. */
		errno	= ENOMEM; return	NULL;	/* insufficient memory */
	}
	outbufp	= memset(iconved_buf, (int)'\0', (outbuf_size + 1));

	/* duplicate the inbuf to dup_inbuf */
	if ((dup_inbuf = malloc((inbuf_length + 1))) == NULL) {
		/* Error: can't allocate duplicated string of `inbuf' */
		free(iconved_buf); iconved_buf = NULL
		;
		errno	= ENOMEM; return	NULL;	/* insufficient memory */
	}
	inbufp	= memcpy(dup_inbuf, inbuf, inbuf_length)
	; dup_inbuf[inbuf_length] = '\0'
	;

	/* sanitize duplicated inbuf (for UTF-8) */
	if (sanitize_for_utf8(inbuf_encode
						, strlen((char *)inbufp)
						, inbufp
						, replace_code
						) == NULL) {
		free(dup_inbuf); dup_inbuf = NULL
		; free(iconved_buf); iconved_buf = NULL
		;
		errno	= ENOMEM; return	NULL;	/* insufficient memory */
	}

	/* create convert-descriptor (EUC-JP or SJIS to UTF-8) */
	tocode	= "UTF-8"
	; fromcode	= ((inbuf_encode == Sanitize_for_UTF8_e_EUCJP)?	"EUC-JP"
					:											"SJIS"
					);
	if ((cd = iconv_open((const char *)tocode
						, (const char *)fromcode
						)) == (iconv_t)(-1)) {
		/* Error: system doesn't support EUC-JP, SJIS or UTF-8 encode. */
		free(dup_inbuf); dup_inbuf = NULL
		; free(iconved_buf); iconved_buf = NULL
		;
		errno	= EPERM; return	NULL; /* system not supported */
	}

	/* (if dup_inbuf isn't empty) convert dup_inbuf to UTF-8 encode. */
	if (((*inbufp != (unsigned char)'\0')
	&& iconv(cd,	(char **)(&inbufp), &inbytesleft
			,		(char **)(&outbufp), &outbytesleft
			) == (size_t)(-1))
	&& (errno == EILSEQ || errno == EINVAL)
	) {	/* Error: `inbuf' contains illegal character (EUC-JP or SJIS) */
		errno	= EILSEQ;	/* found illegal sequence in `inbuf' */
	}
	else { /* process the UTF-8 encode string (iconved_buf) */
		unsigned char	*inbp = iconved_buf, *outbp = outbuf;

		for (result = outbuf, outbytesleft = outbuf_size
		; (*inbp != (unsigned char)'\0') && outbytesleft > 0
		;
		) {
			size_t	move_size;

			/* replace Yen-sign to UTF-8's 'Yen' code ("\xc2\xa5"),
			 * or replace Yen-sign to UTF-8's `backslash' code ("\x5c")
			 */
			if (((how & Sanitize_for_UTF8_Yen_sign_must_be_U005C) == 0)
			&& *inbp == (unsigned char)'\\') {
				if (outbytesleft < 2) {
					break; /* outbuf is insufficient left size */
				}
				memcpy(outbp, "\xc2\xa5", 2)
				; ++inbp; outbp += 2; outbytesleft -= 2
				; continue
				;
			}
			if (((how & Sanitize_for_UTF8_Yen_sign_must_be_U005C) != 0)
			&& memcmp(inbp, "\xc2\xa5", 2) == 0) {
				*outbp	= (unsigned char)'\\'
				; inbp += 2; ++outbp; --outbytesleft
				; continue
				;
			}

			/* replace UTF-8's tilde (3 bytes code) to ASCII's ('\x7e') */
			if (((how & Sanitize_for_UTF8_Tilde_must_be_U203E) == 0)
			&& memcmp(inbp, "\xe2\x80\xbe", 3) == 0) {
				*outbp	= '\x7e'
				; inbp += 3; ++outbp; --outbytesleft
				; continue
				;
			}
			if (((how & Sanitize_for_UTF8_Tilde_must_be_U203E) != 0)
			&& *inbp == (unsigned char)'\x7e') {
				if (outbytesleft < 3) {
					break; /* outbuf is insufficient left size */
				}
				memcpy(outbp, "\xe2\x80\xbe", 3)
				; ++inbp ; outbp += 3; outbytesleft -= 3
				; continue
				;
			}

			/* get the each size of UTF-8 characters (1, 2, 3 or 4 bytes) */
			if ((*inbp & (unsigned char)0x80) == (unsigned char)0)
			{	move_size	= 1;	}
			else if ((*inbp & (unsigned char)0xe0) == (unsigned char)0xc0)
			{	move_size	= 2;	}
			else if ((*inbp & (unsigned char)0xf0) == (unsigned char)0xe0)
			{	move_size	= 3;	}
			else if ((*inbp & (unsigned char)0xf8) == (unsigned char)0xf0)
			{	move_size	= 4;	}
			else if ((*inbp & (unsigned char)0xfc) == (unsigned char)0xf8)
			{	move_size	= 5;	}
			else if ((*inbp & (unsigned char)0xfe) == (unsigned char)0xfc)
			{	move_size	= 6;	}
			else { /* Error: found illegal sequence in UTF-8 string */
				errno = EILSEQ; result = (unsigned char *)NULL
				; break	/* illegal sequence */
				;
			}
			
			/* transfer a UTF-8 character to outbuf */
			if (outbytesleft < move_size) {
				break; /* insufficient left size of outbuf */
			}
			memcpy(outbp, inbp, move_size)
			; inbp += move_size
			; outbp += move_size; outbytesleft -= move_size
			;
		}

		/* (if outbuf has left space) fill padding_code to outbuf */
		if (outbytesleft > 0) {
			memset(outbp, (int)padding_code, outbytesleft);
		}
	}

	/* release convert-descriptor and allocated memories */
	sv_errno	= errno
	; (void)iconv_close(cd); cd = (iconv_t)(-1)
	; free(dup_inbuf); dup_inbuf = NULL
	; free(iconved_buf); iconved_buf = NULL
	; errno	= sv_errno
	;

	/* if error occurred, returns NULL. otherwise, returns address of outbuf
	 */
	return	result;
}

/* convert the UTF-8 encoded string to EUC-JP or SJIS encoded.
 * EINVAL means "illegal arguments".
 * EPERM means "system not support EUC-JP, SJIS or UTF-8 encode".
 * EILSEQ means "`inbuf' contains illegal character".
 */
unsigned char *from_utf8(Sanitize_for_UTF8_e	outbuf_encode
						, unsigned char			*inbuf
						, size_t				inbuf_length
						, unsigned char			*outbuf
						, size_t				outbuf_size
						, char					padding_code
						)
{
	unsigned char	*result = NULL, *reduced_inbuf = NULL, *inbufp = NULL
					, *outbufp = NULL
					;
	iconv_t			cd = (iconv_t)(-1);
	size_t			inbytesleft = inbuf_length, outbytesleft = outbuf_size;
	char			*fromcode = NULL, *tocode = NULL;
	int				sv_errno = 0;

	if (((((outbuf_encode < Sanitize_for_UTF8_e_EUCJP)
	|| outbuf_encode >= Sanitize_for_UTF8_e_N)
	|| inbuf == NULL)
	|| outbuf == NULL)
	) {	/* Error: argument is illegal. */
		errno	= EINVAL;	return	NULL; /* invalid arguments */
	}

	/* allocate the buffer of reduced `inbuf' */
	if ((reduced_inbuf = malloc((inbuf_length + 1))) == NULL) {
		/* Error: can't allocate duplicated buffer of `inbuf'. */
		errno	= ENOMEM; return	NULL;	/* insufficient memory */
	}
	outbufp	= memset(reduced_inbuf, (int)'\0', (inbuf_length + 1));

	/* reduce redundant UTF-8 string. */
	for (inbufp = inbuf
	; (*inbuf != (unsigned char)'\0') && inbytesleft > 0
	;
	) {
		if (*inbufp < (unsigned char)0x80) {
			*outbufp	= *inbufp
			; ++outbufp; ++inbufp; --inbytesleft
			; continue
			;
		}
		else if (*inbufp < (unsigned char)0xc2) {
			if (inbytesleft < 2) {
				goto Illegal_sequence;
			}
Reduce_1_byte:
			*outbufp = ((0x40 & (*inbufp << 6)) | (*(inbufp + 1) & 0x3f))
			; ++outbufp; inbufp += 2; inbytesleft -= 2
			; continue
			;
		}
		else if ((*inbufp & (unsigned char)0xe0) == (unsigned char)0xc0) {
			if (inbytesleft < 2) {
				goto Illegal_sequence;
			}
			memcpy((char *)outbufp, inbufp, 2)
			; outbufp += 2; inbufp += 2; inbytesleft -= 2
			; continue
			;
		}
		else if (*inbufp == (unsigned char)0xe0) {
			if (inbytesleft < 3) {
				goto Illegal_sequence;
			}
			if (*(inbufp + 1) >= (unsigned char)0xa0) {
Transfer_3_bytes:
				memcpy((char *)outbufp, inbufp, 3)
				; outbufp += 3; inbufp += 3; inbytesleft -= 3
				; continue
				;
			}

Reduce_2_bytes:
			if (*(inbufp + 1) >= (unsigned char)0x82) {
				*outbufp	= (0xc0 | (*(inbufp + 1) & 0x1f))
				; *(outbufp + 1)	= *(inbufp + 2)
				; outbufp += 2; inbufp += 3; inbytesleft -= 3
				; continue
				;
			}
			else {
				++inbufp; --inbytesleft; goto Reduce_1_byte;
			}
		}
		else if ((*inbufp & 0xf0) == (unsigned char)0xe0) {
			if (inbytesleft < 3) {
				goto Illegal_sequence;
			}
			goto Transfer_3_bytes;
		}
		else if (*inbufp == (unsigned char)0xf0) {
			if (inbytesleft < 4) {
				goto Illegal_sequence;
			}
			if (*(inbufp + 1) >= (unsigned char)0x90) {
Transfer_4_bytes:
				memcpy((char *)outbufp, inbufp, 4)
				; outbufp += 4; inbufp += 4; inbytesleft -= 4
				; continue
				;
			}

Reduce_3_bytes:
			if ((*(inbufp + 1) >= (unsigned char)0x81)
			|| ((*(inbufp + 1) == (unsigned char)0x80)
				&& *(inbufp + 2) >= (unsigned char)0xa0
				)
			) {
				*outbufp	= (0xe0 | (*(inbufp + 1) & 0x0f))
				; memcpy((char *)(outbufp + 1), (inbufp + 2), 2)
				; outbufp += 3; inbufp += 4; inbytesleft -= 4
				; continue
				;
			}
			else {
				++inbufp; --inbytesleft; goto Reduce_2_bytes;
			}
		}
		else if ((*inbufp & 0xf8) == (unsigned char)0xf0) {
			if (inbytesleft < 4) {
				goto Illegal_sequence;
			}
			goto Transfer_4_bytes;
		}
		else if (*inbufp == (unsigned char)0xf8) {
			if (inbytesleft < 5) {
				goto Illegal_sequence;
			}
			if (*(inbufp + 1) >= (unsigned char)0x88) {
Transfer_5_bytes:
				memcpy((char *)outbuf, inbufp, 5)
				; outbufp += 5; inbufp += 5; inbytesleft -= 5
				; continue
				;
			}

Reduce_4_bytes:
			if (*(inbufp + 1) >= (unsigned char)0x81) {
				*outbufp	= (0xf0 | (*(inbufp + 1) & 0x07))
				; memcpy((char *)(outbufp + 1), (inbufp + 2), 3)
				; outbufp += 4; inbufp += 5; inbytesleft -= 5
				; continue
				;
			}
			else {
				++inbufp; --inbytesleft; goto Reduce_3_bytes;
			}
		}
		else if ((*inbufp & 0xfc) == (unsigned char)0xf8) {
			if (inbytesleft < 5) {
				goto Illegal_sequence;
			}
			goto Transfer_5_bytes;
		}
		else if (*inbufp == (unsigned char)0xfc) {
			if (inbytesleft < 6) {
				goto Illegal_sequence;
			}
			if (*(inbufp + 1) >= (unsigned char)0x84) {
Transfer_6_bytes:
				memcpy((char *)outbufp, inbufp, 6)
				; outbufp += 6; inbufp += 6; inbytesleft -= 6
				; continue
				;
			}

			if (*(inbufp + 1) >= (unsigned char)0x81) {
				*outbufp	= (0xf8 | (*(inbufp + 1) & 0x03))
				; memcpy((char *)(outbufp + 1), (inbufp + 2), 4)
				; outbufp += 5; inbufp += 6; inbytesleft -= 6
				; continue
				;
			}
			else {
				++inbufp; --inbytesleft; goto Reduce_4_bytes;
			}
		}
		else if ((*inbufp & 0xfe) == (unsigned char)0xfc) {
			if (inbytesleft < 6) {
				goto Illegal_sequence;
			}
			goto Transfer_6_bytes;
		}
		else {
Illegal_sequence: /* Error: inbuf contains illegal sequence */
			free(reduced_inbuf); reduced_inbuf = NULL
			;
			errno	= EILSEQ; return NULL; /* illegal sequence */
		}
	}

	/* get length of the reduced UTF-8 string (inbytesleft), inbufp,
	 * outbytesleft and outbufp
	 */
	inbytesleft		= (size_t)(outbufp - reduced_inbuf)
	; inbufp		= reduced_inbuf
	; outbytesleft	= outbuf_size
	; outbufp		= outbuf
	;

	/* create convert-descriptor (EUC-JP or SJIS to UTF-8) */
	fromcode	= "UTF-8"
	; tocode	= ((outbuf_encode == Sanitize_for_UTF8_e_EUCJP)?	"EUC-JP"
					:												"SJIS"
					);
	if ((cd = iconv_open((const char *)tocode
						, (const char *)fromcode
						)) == (iconv_t)(-1)) {
		/* Error: system doesn't support EUC-JP, SJIS or UTF-8 encode. */
		free(reduced_inbuf); reduced_inbuf = NULL
		;
		errno	= EPERM; return	NULL; /* system not supported */
	}

	/* (if dup_inbuf isn't empty) convert dup_inbuf to UTF-8 encode. */
	if (((*inbufp != (unsigned char)'\0')
	&& iconv(cd,	(char **)(&inbufp), &inbytesleft
			,		(char **)(&outbufp), &outbytesleft
			) == (size_t)(-1))
	&& (errno == EILSEQ || errno == EINVAL)
	) {	/* Error: `inbuf' contains illegal character (EUC-JP or SJIS) */
		errno	= EILSEQ;	/* found illegal sequence in `inbuf' */
	}
	else {
		/* (if outbuf has left space) fill padding_code to outbuf */
		if (outbytesleft > 0) {
			memset(outbufp, (int)padding_code, outbytesleft);
		}

		result	= outbuf;
	}

	/* release convert-descriptor and allocated buffer */
	sv_errno	= errno
	; (void)iconv_close(cd); cd = (iconv_t)(-1)
	; free(reduced_inbuf); reduced_inbuf = NULL
	; errno	= sv_errno
	;

	/* if error occurred, returns NULL. otherwise, returns address of outbuf
	 */
	return	result;
}
